/**
*** step_theme_main.c:	ColorStep initialization code.
***
*** Written by:		Ullrich Hafner
***    	
*** Copyright (C) 1999 Ullrich Hafner <hafner@bigfoot.de>
***
*** Modified by:        Juan Pablo Romero Mndez <juan_pablo_22@yahoo.com>
*** Modified by:        Pierre Baillargeon
***
*** This program is free software; you can redistribute it and/or modify
*** it under the terms of the GNU General Public License as published by
*** the Free Software Foundation; either version 2 of the License, or
*** (at your option) any later version.
***
*** This program is distributed in the hope that it will be useful,
*** but WITHOUT ANY WARRANTY; without even the implied warranty of
*** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*** GNU General Public License for more details.
***
*** You should have received a copy of the GNU General Public License
*** along with this program; if not, write to the Free Software
*** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
**/


#include <gtk/gtk.h>
#include <gmodule.h>
#include "step_theme_main.h"

#define SCROLLBAR_WIDTH 16
#define MINSLIDER_WIDTH 25
#define INDICATOR_SIZE	16
#define SCALE_WIDTH     16
#define LIGHTNESS_MULT  1.25
#define DARKNESS_MULT   0.6

/* from gtktearoffmenuitem.c */
#define TEAR_LENGTH 3
#define BORDER_SPACING 0
#define MIN_TEXT_WIDTH_LINES     20
#define MIN_TEXT_HEIGHT_LINES    10


/***************************************************************************
****************************************************************************
***
*** Theme special colors.
**/

GdkGC *colorstep_darker_gc;
GdkGC *colorstep_darker2_gc;
GdkGC *colorstep_blue_darker_gc;


/***************************************************************************
****************************************************************************
***
*** Nicked from Xenophilia
**/

static void
step_vscrollbar_size_allocate (GtkWidget *widget, GtkAllocation *allocation);
static void
step_hscrollbar_size_allocate (GtkWidget *widget, GtkAllocation *allocation);
static void
step_vscrollbar_realize (GtkWidget *widget);
static void
step_hscrollbar_realize (GtkWidget *widget);
static void
step_vscrollbar_slider_update (GtkRange *range);
static void
step_hscrollbar_slider_update (GtkRange *range);
static gint
step_vscrollbar_button_press (GtkWidget *widget, GdkEventButton *event);
static gint
step_hscrollbar_button_press (GtkWidget *widget, GdkEventButton *event);
static void
step_vpaned_realize (GtkWidget *);
static void
step_hpaned_realize (GtkWidget *);
static void
step_vpaned_size_allocate(GtkWidget *, GtkAllocation *);
static void
step_hpaned_size_allocate(GtkWidget *, GtkAllocation *);

void (* old_vscrollbar_realize)(GtkWidget *);
void (* old_hscrollbar_realize)(GtkWidget *);
void (* old_vscrollbar_slider_update)(GtkRange *);
void (* old_hscrollbar_slider_update)(GtkRange *);
void (* old_vpaned_size_allocate)(GtkWidget *, GtkAllocation *);
void (* old_hpaned_size_allocate)(GtkWidget *, GtkAllocation *);
void (* old_vpaned_realize)(GtkWidget *);
void (* old_hpaned_realize)(GtkWidget *);


/***************************************************************************
****************************************************************************
***
*** color stuff
**/

static void
step_style_shade (GdkColor *a,
		  GdkColor *b,
		  gdouble   k);
static void
hls_to_rgb (gdouble *h,
	    gdouble *l,
	    gdouble *s);
static void
rgb_to_hls (gdouble *r,
	    gdouble *g,
	    gdouble *b);

static void
step_style_shade (GdkColor *a,
		  GdkColor *b,
		  gdouble   k)
{
  gdouble red;
  gdouble green;
  gdouble blue;

  red = (gdouble) a->red / 65535.0;
  green = (gdouble) a->green / 65535.0;
  blue = (gdouble) a->blue / 65535.0;

  rgb_to_hls (&red, &green, &blue);

  green *= k;
  if (green > 1.0)
    green = 1.0;
  else if (green < 0.0)
    green = 0.0;

  blue *= k;
  if (blue > 1.0)
    blue = 1.0;
  else if (blue < 0.0)
    blue = 0.0;

  hls_to_rgb (&red, &green, &blue);

  b->red = red * 65535.0;
  b->green = green * 65535.0;
  b->blue = blue * 65535.0;
}

static void
hls_to_rgb (gdouble *h,
	    gdouble *l,
	    gdouble *s)
{
  gdouble hue;
  gdouble lightness;
  gdouble saturation;
  gdouble m1, m2;
  gdouble r, g, b;

  lightness = *l;
  saturation = *s;

  if (lightness <= 0.5)
    m2 = lightness * (1 + saturation);
  else
    m2 = lightness + saturation - lightness * saturation;
  m1 = 2 * lightness - m2;

  if (saturation == 0)
    {
      *h = lightness;
      *l = lightness;
      *s = lightness;
    }
  else
    {
      hue = *h + 120;
      while (hue > 360)
	hue -= 360;
      while (hue < 0)
	hue += 360;

      if (hue < 60)
	r = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
	r = m2;
      else if (hue < 240)
	r = m1 + (m2 - m1) * (240 - hue) / 60;
      else
	r = m1;

      hue = *h;
      while (hue > 360)
	hue -= 360;
      while (hue < 0)
	hue += 360;

      if (hue < 60)
	g = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
	g = m2;
      else if (hue < 240)
	g = m1 + (m2 - m1) * (240 - hue) / 60;
      else
	g = m1;

      hue = *h - 120;
      while (hue > 360)
	hue -= 360;
      while (hue < 0)
	hue += 360;

      if (hue < 60)
	b = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
	b = m2;
      else if (hue < 240)
	b = m1 + (m2 - m1) * (240 - hue) / 60;
      else
	b = m1;

      *h = r;
      *l = g;
      *s = b;
    }
}

static void
rgb_to_hls (gdouble *r,
	    gdouble *g,
	    gdouble *b)
{
  gdouble min;
  gdouble max;
  gdouble red;
  gdouble green;
  gdouble blue;
  gdouble h, l, s;
  gdouble delta;

  red = *r;
  green = *g;
  blue = *b;

  if (red > green)
    {
      if (red > blue)
	max = red;
      else
	max = blue;

      if (green < blue)
	min = green;
      else
	min = blue;
    }
  else
    {
      if (green > blue)
	max = green;
      else
	max = blue;

      if (red < blue)
	min = red;
      else
	min = blue;
    }

  l = (max + min) / 2;
  s = 0;
  h = 0;

  if (max != min)
    {
      if (l <= 0.5)
	s = (max - min) / (max + min);
      else
	s = (max - min) / (2 - max - min);

      delta = max -min;
      if (red == max)
	h = (green - blue) / delta;
      else if (green == max)
	h = 2 + (blue - red) / delta;
      else if (blue == max)
	h = 4 + (red - green) / delta;

      h *= 60;
      if (h < 0.0)
	h += 360;
    }

  *r = h;
  *g = l;
  *b = s;
}


/***************************************************************************
****************************************************************************
***
*** Theme engine indirect functions implementation.
**/

static guint
step_parse_rc_style (GScanner *scanner, GtkRcStyle *rc_style)
{
   guint token;

   token = g_scanner_peek_next_token (scanner);
   while (token != G_TOKEN_RIGHT_CURLY)
   {
      switch (token)
      {
	 default:
	    g_scanner_get_next_token (scanner);
	    token = G_TOKEN_RIGHT_CURLY;
	    break;
      }

      if (token != G_TOKEN_NONE)
	 return token;

      token = g_scanner_peek_next_token (scanner);
   }

   g_scanner_get_next_token (scanner);

   rc_style->engine_data = NULL;

   return G_TOKEN_NONE;
}

static void
step_merge_rc_style (GtkRcStyle *dest, GtkRcStyle *src)
{
}

static void
step_rc_style_to_style (GtkStyle *style, GtkRcStyle *rc_style)
{
   style->klass = &step_class;
}

static void
step_duplicate_style (GtkStyle *dest, GtkStyle *src)
{
}

static void
step_realize_style (GtkStyle *style)
{
    GdkGCValuesMask gc_values_mask = GDK_GC_FOREGROUND | GDK_GC_FONT;
    GdkGCValues	   gc_values;
    gint		   i;
    extern GdkFont  *default_font;

    if ( !default_font )
        default_font = gdk_font_load ("-*-helvetica-medium-r-normal--*-120-*-*-*-*-*-*");

    if (style->font->type == GDK_FONT_FONT)
        gc_values.font = style->font;
    else if (style->font->type == GDK_FONT_FONTSET)
        gc_values.font = default_font;

    for (i = 0; i < 5; i++)
    {
        /*
         *  Release old GC
         */
        gtk_gc_release (style->light_gc [i]);
        gtk_gc_release (style->dark_gc [i]);

        /*
         *  Compute new colors
         */
        step_style_shade (&style->bg [i], &style->light [i], LIGHTNESS_MULT);
        step_style_shade (&style->bg [i], &style->dark [i], DARKNESS_MULT);

        style->mid [i].red = (style->light [i].red + style->dark [i].red) / 2;
        style->mid [i].green = (style->light [i].green + style->dark [i].green) / 2;
        style->mid [i].blue = (style->light [i].blue + style->dark [i].blue) / 2;

        /*
         *  Allocate new gc
         */
        if (!gdk_color_alloc (style->colormap, &style->light [i]))
            g_warning ("unable to allocate color: ( %d %d %d )",
                       style->light [i].red, style->light [i].green,
                       style->light [i].blue);
        if (!gdk_color_alloc (style->colormap, &style->dark [i]))
            g_warning ("unable to allocate color: ( %d %d %d )",
                       style->dark [i].red, style->dark [i].green,
                       style->dark [i].blue);
        if (!gdk_color_alloc (style->colormap, &style->mid [i]))
            g_warning ("unable to allocate color: ( %d %d %d )",
                       style->mid [i].red, style->mid [i].green,
                       style->mid [i].blue);

        gc_values.foreground = style->light [i];
        style->light_gc [i] = gtk_gc_get (style->depth, style->colormap,
                                          &gc_values, gc_values_mask);

        gc_values.foreground = style->dark[i];
        style->dark_gc [i] = gtk_gc_get (style->depth, style->colormap,
                                         &gc_values, gc_values_mask);

        gc_values.foreground = style->mid [i];
        style->mid_gc [i] = gtk_gc_get (style->depth, style->colormap,
                                        &gc_values, gc_values_mask);
    }

    {
       static int done = 0;

       if ( ! done )
       {
	  static char radio_s_on_buf[32];
	  static char radio_s_off_buf[32];
	  static char radio_on_prelight_buf[32];
	  static char radio_on_prelight_buf2[32];
	  static char radio_on_buf[32];
	  static char radio_off_prelight_buf[32];
	  static char radio_off_prelight_buf2[32];
	  static char radio_off_buf[32];

	  sprintf ( radio_s_on_buf, "c c #%02.2x%02.2x%02.2x",
#ifdef ALWAYS_PRELIGHT_SELECTION
		    style->bg [GTK_STATE_PRELIGHT].red / 256,
		    style->bg [GTK_STATE_PRELIGHT].green / 256,
		    style->bg [GTK_STATE_PRELIGHT].blue / 256
#else
		    style->bg [GTK_STATE_ACTIVE].red / 256,
		    style->bg [GTK_STATE_ACTIVE].green / 256,
		    style->bg [GTK_STATE_ACTIVE].blue / 256
#endif	  
		  );
	  sprintf ( radio_s_off_buf, "c c #%02.2x%02.2x%02.2x",
		    style->bg [GTK_STATE_ACTIVE].red / 256,
		    style->bg [GTK_STATE_ACTIVE].green / 256,
		    style->bg [GTK_STATE_ACTIVE].blue / 256 );
	  sprintf ( radio_on_prelight_buf, "c c #%02.2x%02.2x%02.2x",
		    style->bg [GTK_STATE_PRELIGHT].red / 256,
		    style->bg [GTK_STATE_PRELIGHT].green / 256,
		    style->bg [GTK_STATE_PRELIGHT].blue / 256 );
	  sprintf ( radio_on_prelight_buf2, "a c #%02.2x%02.2x%02.2x",
		    style->light [GTK_STATE_PRELIGHT].red / 256,
		    style->light [GTK_STATE_PRELIGHT].green / 256,
		    style->light [GTK_STATE_PRELIGHT].blue / 256 );
	  sprintf ( radio_on_buf, "c c #%02.2x%02.2x%02.2x",
#ifdef ALWAYS_PRELIGHT_SELECTION
		    style->bg [GTK_STATE_PRELIGHT].red / 256,
		    style->bg [GTK_STATE_PRELIGHT].green / 256,
		    style->bg [GTK_STATE_PRELIGHT].blue / 256
#else
		    style->bg [GTK_STATE_ACTIVE].red / 256,
		    style->bg [GTK_STATE_ACTIVE].green / 256,
		    style->bg [GTK_STATE_ACTIVE].blue / 256
#endif	  
		  );
	  sprintf ( radio_off_prelight_buf, "c c #%02.2x%02.2x%02.2x",
		    style->bg [GTK_STATE_PRELIGHT].red / 256,
		    style->bg [GTK_STATE_PRELIGHT].green / 256,
		    style->bg [GTK_STATE_PRELIGHT].blue / 256 );
	  sprintf ( radio_off_prelight_buf2, "a c #%02.2x%02.2x%02.2x",
		    style->light [GTK_STATE_PRELIGHT].red / 256,
		    style->light [GTK_STATE_PRELIGHT].green / 256,
		    style->light [GTK_STATE_PRELIGHT].blue / 256 );
	  sprintf ( radio_off_buf, "c c #%02.2x%02.2x%02.2x",
		    style->bg [GTK_STATE_ACTIVE].red / 256,
		    style->bg [GTK_STATE_ACTIVE].green / 256,
		    style->bg [GTK_STATE_ACTIVE].blue / 256 );

	  color_step_radio_s_on [6]         = radio_s_on_buf;
	  color_step_radio_s_off [6]        = radio_s_off_buf;
	  color_step_radio_on_prelight [6]  = radio_on_prelight_buf;
	  color_step_radio_on_prelight [4]  = radio_on_prelight_buf2;
	  color_step_radio_on [6]           = radio_on_buf;
	  color_step_radio_off_prelight [6] = radio_off_prelight_buf;
	  color_step_radio_off_prelight [4] = radio_off_prelight_buf2;
	  color_step_radio_off [6]          = radio_off_buf;

	  done = 1;
       }
    }

    {
        GdkGCValues values;
        GdkColormap *colormap;
        gint depth;
        GdkColor shade;

        colormap = gdk_colormap_get_system();
        depth = gdk_visual_get_system()->depth;

        step_style_shade (&style->dark[0], &shade, 0.85);
        if (!gdk_color_alloc (colormap, &shade)) {
            g_warning ("unable to allocate color: ( %d %d %d )",
                       shade.red, shade.green, shade.blue);
        }
        values.foreground = shade;
        colorstep_darker_gc = gtk_gc_get(depth, colormap, &values, GDK_GC_FOREGROUND);

        step_style_shade (&style->dark[0], &shade, 0.6);
        if (!gdk_color_alloc (colormap, &shade)) {
            g_warning ("unable to allocate color: ( %d %d %d )",
                       shade.red, shade.green, shade.blue);
        }
        values.foreground = shade;
        colorstep_darker2_gc = gtk_gc_get(depth, colormap, &values, GDK_GC_FOREGROUND);

        step_style_shade (&style->dark[0], &shade, 0.7);
        if (!gdk_color_alloc (colormap, &shade)) {
            g_warning ("unable to allocate color: ( %d %d %d )",
                       shade.red, shade.green, shade.blue);
        }
        values.foreground = shade;
        colorstep_blue_darker_gc = gtk_gc_get(depth, colormap, &values, GDK_GC_FOREGROUND);
    }
}

static void
step_unrealize_style (GtkStyle *style)
{
}

static void
step_destroy_rc_style (GtkRcStyle *rc_style)
{
}

static void
step_destroy_style (GtkStyle *style)
{
}

static void step_tearoff_menu_item_draw(GtkWidget *widget, GdkRectangle *area);
static gint step_tearoff_menu_item_expose(GtkWidget *widget, GdkEventExpose *event);
static void step_tearoff_menu_item_size_request (GtkWidget *widget, GtkRequisition *requisition);


/***************************************************************************
****************************************************************************
***
*** Theme engine entry points.
**/

void theme_init (GtkThemeEngine *engine)
{
    engine->parse_rc_style    = step_parse_rc_style;
    engine->merge_rc_style    = step_merge_rc_style;
    engine->rc_style_to_style = step_rc_style_to_style;
    engine->duplicate_style   = step_duplicate_style;
    engine->realize_style     = step_realize_style;
    engine->unrealize_style   = step_unrealize_style;
    engine->destroy_rc_style  = step_destroy_rc_style;
    engine->destroy_style     = step_destroy_style;
    engine->set_background    = NULL;

    /*
     *  Make scrollbars wider
     */
    {
        GtkRangeClass	*rangeclass;
        GtkScrollbarClass *scrollbarclass;

        scrollbarclass
            = (GtkScrollbarClass *) gtk_type_class (GTK_TYPE_SCROLLBAR);
        rangeclass = (GtkRangeClass *) scrollbarclass;

        rangeclass->slider_width           = SCROLLBAR_WIDTH;
        rangeclass->min_slider_size        = MINSLIDER_WIDTH;
        rangeclass->stepper_size           = SCROLLBAR_WIDTH;
        rangeclass->stepper_slider_spacing = 1;
    }

    /* scrollbar spacing */
    {
        GtkScrolledWindowClass *scrolledclass;
        scrolledclass = (GtkScrolledWindowClass *) gtk_type_class (GTK_TYPE_SCROLLED_WINDOW);
        scrolledclass->scrollbar_spacing = 0;
    }

    /* tearoff menus  */
    {
        GtkTearoffMenuItemClass* tearoffclass;
        GtkWidgetClass *widget_class;

        tearoffclass = (GtkTearoffMenuItemClass *) gtk_type_class (GTK_TYPE_TEAROFF_MENU_ITEM);
        widget_class = (GtkWidgetClass*) tearoffclass;

        widget_class->draw = step_tearoff_menu_item_draw;
        widget_class->expose_event = step_tearoff_menu_item_expose;
        widget_class->size_request = step_tearoff_menu_item_size_request;
    }


    { /* text widget */
        /*
        GtkTextClass* textclass;
        GtkWidgetClass *widget_class;

        textclass = (GtkTextClass*) gtk_type_class(gtk_text_get_type());
        widget_class = (GtkWidgetClass *) textclass;


        widget_class->realize = step_text_realize;
        widget_class->draw_focus = step_text_draw_focus;
        widget_class->size_request = step_text_size_request;
        widget_class->size_allocate = step_text_size_allocate;
       */
    }


    /*
     *  Move the scrollbar buttons to the lower right corner
     */
    {
        GtkWidgetClass	*widget_class;
        GtkRangeClass	*range_class;

        widget_class
            = (GtkWidgetClass *) gtk_type_class (GTK_TYPE_VSCROLLBAR);
        range_class 		    = (GtkRangeClass *) widget_class;
        old_vscrollbar_realize	    = widget_class->realize;
        old_vscrollbar_slider_update = range_class->slider_update;

        widget_class->size_allocate  	= step_vscrollbar_size_allocate;
        widget_class->realize		= step_vscrollbar_realize;
        widget_class->button_press_event = step_vscrollbar_button_press;
        range_class->slider_update	= step_vscrollbar_slider_update;

        widget_class
            = (GtkWidgetClass *) gtk_type_class (GTK_TYPE_HSCROLLBAR);
        range_class 		    = (GtkRangeClass *) widget_class;
        old_hscrollbar_realize	    = widget_class->realize;
        old_hscrollbar_slider_update = range_class->slider_update;

        widget_class->size_allocate  	= step_hscrollbar_size_allocate;
        widget_class->realize		= step_hscrollbar_realize;
        widget_class->button_press_event = step_hscrollbar_button_press;
        range_class->slider_update 	= step_hscrollbar_slider_update;
    }

    /*
     *  Make radio and check buttons larger
     */
    {
        GtkRadioButtonClass *radiobuttonclass;
        GtkCheckButtonClass *checkbuttonclass;

        radiobuttonclass
            = (GtkRadioButtonClass *) gtk_type_class (GTK_TYPE_RADIO_BUTTON);
        checkbuttonclass = (GtkCheckButtonClass *) radiobuttonclass;
        checkbuttonclass->indicator_size = INDICATOR_SIZE;

        checkbuttonclass
            = (GtkCheckButtonClass *) gtk_type_class (GTK_TYPE_CHECK_BUTTON);
        checkbuttonclass->indicator_size = INDICATOR_SIZE;
    }

    /*
     *  Make the reisze handle of panes be full size to ease movements.
     */
    {
        GtkWidgetClass	*widget_class;

        widget_class
            = (GtkWidgetClass *) gtk_type_class (gtk_vpaned_get_type());
        old_vpaned_realize	    = widget_class->realize;
        old_vpaned_size_allocate = widget_class->size_allocate;

        widget_class->realize		= step_vpaned_realize;
        widget_class->size_allocate	= step_vpaned_size_allocate;

        widget_class
            = (GtkWidgetClass *) gtk_type_class (gtk_hpaned_get_type());
        old_hpaned_realize	    = widget_class->realize;
        old_hpaned_size_allocate = widget_class->size_allocate;

        widget_class->realize		= step_hpaned_realize;
        widget_class->size_allocate 	= step_hpaned_size_allocate;
    }

}

void
theme_exit (void)
{
}

/*
 * The following function will be called by GTK+ when the module
 * is loaded and checks to see if we are compatible with the
 * version of GTK+ that loads us.
 */

G_MODULE_EXPORT const gchar*
g_module_check_init (GModule *module);

const gchar*
g_module_check_init (GModule *module)
{
   return gtk_check_version (GTK_MAJOR_VERSION,
			     GTK_MINOR_VERSION,
			     GTK_MICRO_VERSION - GTK_INTERFACE_AGE);
}


/***************************************************************************
****************************************************************************
***
*** tearoff menus.
**/

static void
step_tearoff_menu_item_size_request (GtkWidget      *widget,
                                     GtkRequisition *requisition)
{
  GtkTearoffMenuItem *tearoff;

  tearoff = GTK_TEAROFF_MENU_ITEM (widget);
  
  requisition->width = (GTK_CONTAINER (widget)->border_width +
			widget->style->klass->xthickness +
			BORDER_SPACING) * 2;
  requisition->height = (GTK_CONTAINER (widget)->border_width +
			 widget->style->klass->ythickness) * 2;

  if (!tearoff->torn_off)
      requisition->height += widget->style->klass->ythickness;
  else
      requisition->height = 0;
}



static void
step_tearoff_menu_item_paint (GtkWidget   *widget,
                              GdkRectangle *area)
{
    GtkMenuItem *menu_item;
    GtkTearoffMenuItem *tearoff_item;
    GtkShadowType shadow_type;
    gint width, height;
    gint x, y;
    gint right_max;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_TEAROFF_MENU_ITEM (widget));

    if (GTK_WIDGET_DRAWABLE (widget))
    {
        menu_item = GTK_MENU_ITEM (widget);
        tearoff_item = GTK_TEAROFF_MENU_ITEM (widget);

        x = GTK_CONTAINER (menu_item)->border_width;
        y = GTK_CONTAINER (menu_item)->border_width;
        width = widget->allocation.width - x * 2;
        height = widget->allocation.height - y * 2;
        right_max = x + width;

        if (widget->state == GTK_STATE_PRELIGHT)
            gtk_paint_box (widget->style,
                           widget->window,
                           GTK_STATE_PRELIGHT,
                           GTK_SHADOW_OUT,
                           area, widget, "menuitem",
                           x, y, width, height);
        else
            gdk_window_clear_area (widget->window, area->x, area->y, area->width, area->height);

        if (!tearoff_item->torn_off)
            while (x < right_max) {
                gtk_draw_hline (widget->style, widget->window, GTK_STATE_NORMAL,
                                x, MIN (x+TEAR_LENGTH, right_max),
                                y + (height - widget->style->klass->ythickness)/2);
                x += 2 * TEAR_LENGTH;
            }
    }
}

static void
step_tearoff_menu_item_draw (GtkWidget    *widget,
			  GdkRectangle *area)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_TEAROFF_MENU_ITEM (widget));
  g_return_if_fail (area != NULL);

  step_tearoff_menu_item_paint (widget, area);
}

static gint
step_tearoff_menu_item_expose (GtkWidget      *widget,
			    GdkEventExpose *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_TEAROFF_MENU_ITEM (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  step_tearoff_menu_item_paint (widget, &event->area);

  return FALSE;
}


/***************************************************************************
****************************************************************************
***
*** Nicked from Xenophilias.
**/

#define RANGE_CLASS(widget) \
	GTK_RANGE_CLASS(GTK_OBJECT(widget)->klass)

static void
step_vscrollbar_realize (GtkWidget *widget)
{
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_VSCROLLBAR (widget));

   old_vscrollbar_realize (widget);
   step_vscrollbar_size_allocate (widget, &(widget->allocation));
}

static void
step_vscrollbar_slider_update (GtkRange *range)
{
   GtkAdjustment *adjustment;
   gint x,y,w,h, ty,th, xt,yt;
   gfloat ah;

   g_return_if_fail (range != NULL);
   g_return_if_fail (GTK_IS_VSCROLLBAR (range));

   if (GTK_WIDGET_REALIZED(GTK_WIDGET(range))) {
      gdk_window_get_geometry (range->trough, NULL, NULL, &w, &th, NULL);
      xt = ((GtkWidget *)range)->style->klass->xthickness;
      yt = ((GtkWidget *)range)->style->klass->ythickness;

      x   = xt;
      y   = yt;
      w  -= xt+xt;
      th -= 2 * (yt + RANGE_CLASS(range)->stepper_size);
      h   = th;

      if (   ((GtkWidget *)range)->parent
	     && GTK_IS_SCROLLED_WINDOW(((GtkWidget *)range)->parent)
	     && (GTK_SCROLLED_WINDOW(((GtkWidget *)range)->parent)->window_placement & 0x01))
      {
	 y += 2 * RANGE_CLASS(range)->stepper_size;
      }

      adjustment = range->adjustment;
      ah = adjustment->upper - adjustment->lower;
		
      if (adjustment->page_size > 0 && ah > 1.0E-5) {
	 if (adjustment->page_size > ah)
	    adjustment->page_size = ah;

	 h = th * range->adjustment->page_size / ah;
	 if (h < RANGE_CLASS(range)->min_slider_size)
	    h = RANGE_CLASS(range)->min_slider_size;
		
	 y += (th-h) * (adjustment->value - adjustment->lower) / (ah - adjustment->page_size);
      }
      if (y<yt) y=yt;
		
      gdk_window_move_resize (range->slider, x, y, w, h);
   }
}


static void
step_vscrollbar_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
   GtkRange *range;
   gint w, s, x, y;
	
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_VSCROLLBAR (widget));
   g_return_if_fail (allocation != NULL);
	
   widget->allocation = *allocation;
   if (GTK_WIDGET_REALIZED (widget)) {
      w = widget->requisition.width - widget->style->klass->xthickness * 2;
      s = RANGE_CLASS(widget)->stepper_size;
      x = widget->style->klass->xthickness;
      y = allocation->height - widget->style->klass->ythickness;
      if (   widget->parent
	     && GTK_IS_SCROLLED_WINDOW(widget->parent)
	     && (GTK_SCROLLED_WINDOW(widget->parent)->window_placement & 0x01))
      {
	 y = widget->style->klass->ythickness + s*2;
      }
      range = GTK_RANGE (widget);
		
      gdk_window_move_resize (range->step_back, x, y - s*2, w, s);
      gdk_window_move_resize (range->step_forw, x, y - s,   w, s);
      gdk_window_move_resize (range->trough,
			      allocation->x + (allocation->width - widget->requisition.width) / 2,
			      allocation->y,
			      widget->allocation.width,
			      allocation->height);

      RANGE_CLASS(range)->slider_update(range);
   }
}

static gint
step_vscrollbar_button_press (GtkWidget *widget, GdkEventButton *event)
{
   GtkRange *range;
   GdkEventButton fake_event;
   gint size;
	
   g_return_val_if_fail (widget != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_VSCROLLBAR (widget), FALSE);
   range = GTK_RANGE(widget);

   fake_event = *event;
   if (range->button == 0) {
      size = RANGE_CLASS(range)->stepper_size;
      if (event->window == range->slider)
	 size = -size;
      if (   widget->parent
	     && GTK_IS_SCROLLED_WINDOW(widget->parent)
	     && (GTK_SCROLLED_WINDOW(widget->parent)->window_placement & 0x01))
      {
	 size = -size;
      }
		
      fake_event.y += size;
   }
   return GTK_WIDGET_CLASS(gtk_type_class(GTK_TYPE_SCROLLBAR))->button_press_event(widget, &fake_event);
}


static void
step_hscrollbar_realize (GtkWidget *widget)
{
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_HSCROLLBAR (widget));

   old_hscrollbar_realize (widget);
   step_hscrollbar_size_allocate (widget, &(widget->allocation));
}

static void
step_hscrollbar_slider_update(GtkRange *range)
{
   GtkAdjustment *adjustment;
   gint x,y,w,h, tx,tw, xt, yt;
   gfloat aw;

   g_return_if_fail (range != NULL);
   g_return_if_fail (GTK_IS_HSCROLLBAR (range));

   if (GTK_WIDGET_REALIZED(GTK_WIDGET(range))) {
      gdk_window_get_geometry (range->trough, NULL, NULL, &tw, &h, NULL);
      xt = ((GtkWidget *)range)->style->klass->xthickness;
      yt = ((GtkWidget *)range)->style->klass->ythickness;

      x   = xt;
      y   = yt;
      h  -= yt+yt;
      tw -= 2 * (xt + RANGE_CLASS(range)->stepper_size);
      w   = tw;

      if (   ((GtkWidget *)range)->parent
	     && GTK_IS_SCROLLED_WINDOW(((GtkWidget *)range)->parent)
	     && (GTK_SCROLLED_WINDOW(((GtkWidget *)range)->parent)->window_placement & 0x02))
      {
	 x += 2 * RANGE_CLASS(range)->stepper_size;
      }

      adjustment = range->adjustment;
      aw = adjustment->upper - adjustment->lower;
		
      if (adjustment->page_size > 0 && aw > 1.0E-5) {
	 if (adjustment->page_size > aw)
	    adjustment->page_size = aw;

	 w = tw * range->adjustment->page_size / aw;
	 if (w < RANGE_CLASS(range)->min_slider_size)
	    w = RANGE_CLASS(range)->min_slider_size;
		
	 x += (tw-w) * (adjustment->value - adjustment->lower) / (aw - adjustment->page_size);
      }
      if (x<xt) x=xt;
		
      gdk_window_move_resize (range->slider, x, y, w, h);
   }
}

static void
step_hscrollbar_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
   GtkRange *range;
   gint h, s, x, y;
	
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_HSCROLLBAR (widget));
   g_return_if_fail (allocation != NULL);
	
   widget->allocation = *allocation;
   if (GTK_WIDGET_REALIZED (widget)) {
      h = widget->requisition.height - widget->style->klass->ythickness * 2;
      s = RANGE_CLASS (widget)->stepper_size;
      y = widget->style->klass->ythickness;
      x = allocation->width - widget->style->klass->xthickness;
      if (   widget->parent
	     && GTK_IS_SCROLLED_WINDOW(widget->parent)
	     && (GTK_SCROLLED_WINDOW(widget->parent)->window_placement & 0x02))
      {
	 x = widget->style->klass->xthickness + s*2;
      }
      range = GTK_RANGE (widget);
		
      gdk_window_move_resize (range->step_back, x - s*2, y, s, h);
      gdk_window_move_resize (range->step_forw, x - s,   y, s, h);
		
      gdk_window_move_resize (range->trough,
			      allocation->x,
			      allocation->y + (allocation->height - widget->requisition.height) / 2,
			      allocation->width, widget->requisition.height);
      RANGE_CLASS(range)->slider_update(range);
   }
}


static gint
step_hscrollbar_button_press (GtkWidget *widget, GdkEventButton *event)
{
   GtkRange *range;
   GdkEventButton fake_event;
	
   g_return_val_if_fail (widget != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_HSCROLLBAR (widget), FALSE);
   range = GTK_RANGE(widget);

   fake_event = *event;
   if (range->button == 0) {
      gint size = RANGE_CLASS(range)->stepper_size;
      if (event->window == range->slider)
	 size = -size;
      if (   widget->parent
	     && GTK_IS_SCROLLED_WINDOW(widget->parent)
	     && (GTK_SCROLLED_WINDOW(widget->parent)->window_placement & 0x02))
      {
	 size = -size;
      }

      fake_event.x += size;
   }
   return GTK_WIDGET_CLASS(gtk_type_class(GTK_TYPE_SCROLLBAR))->
       button_press_event(widget, &fake_event);
}


/* Panes */
static void step_hpaned_realize (GtkWidget *widget) {
	GdkCursor *crsr;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_HPANED(widget));

	old_hpaned_realize(widget);
	crsr = gdk_cursor_new(GDK_SB_H_DOUBLE_ARROW);
	gdk_window_set_cursor(GTK_PANED(widget)->handle, crsr);
	gdk_cursor_destroy(crsr);
	
	step_hpaned_size_allocate (widget, &(widget->allocation));
}

static void step_hpaned_size_allocate (GtkWidget *widget, GtkAllocation *allocation) {
	GtkPaned	*paned;
	gint		x, width;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_HPANED(widget));

	paned = ((GtkPaned*)widget);
	paned->handle_size = paned->gutter_size;

	old_hpaned_size_allocate (widget, allocation);
	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_get_geometry(paned->handle, &x, NULL, &width, NULL, NULL);
		gdk_window_move_resize (paned->handle,
								x,
								0,
								paned->handle_size,
								widget->allocation.height);
	}
	gtk_widget_queue_draw(widget);
}

static void step_vpaned_realize (GtkWidget *widget) {
	GdkCursor *crsr;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_VPANED(widget));

	old_vpaned_realize(widget);
	crsr = gdk_cursor_new(GDK_SB_V_DOUBLE_ARROW);
	gdk_window_set_cursor(GTK_PANED(widget)->handle, crsr);
	gdk_cursor_destroy(crsr);
	step_vpaned_size_allocate (widget, &(widget->allocation));
}

static void step_vpaned_size_allocate (GtkWidget *widget, GtkAllocation *allocation) {
	GtkPaned	*paned;
	gint		y, height;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_VPANED(widget));

	paned = ((GtkPaned*)widget);
	paned->handle_size = paned->gutter_size;

	old_vpaned_size_allocate (widget, allocation);
	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_get_geometry(paned->handle, NULL, &y, NULL, &height, NULL);
		gdk_window_move_resize (paned->handle,
								0,
								y,
								widget->allocation.width,
								paned->handle_size);
	}
	gtk_widget_queue_draw(widget);
}
